from typing import List, Optional

from pydantic import BaseModel, Field

from deployment.bff.app.v1.models.helper import (
    BaseIndexRequestModel,
    BaseSearchResponseModel,
)


# Request Model
class NowVideoIndexRequestModel(BaseIndexRequestModel):
    videos: Optional[List[str]] = Field(
        default=...,
        description='List of Videos to index. Videos should be base64encoded in `utf-8` format',
    )


# Response Model
class NowVideoResponseModel(BaseSearchResponseModel):
    blob: Optional[str] = Field(
        description='Base64 encoded video in `utf-8` str format.'
    )


class NowVideoListResponseModel(BaseModel):
    __root__: List[NowVideoResponseModel] = Field(description='list of video responses')


NowVideoResponseModel.update_forward_refs()



"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.
"""
from __future__ import annotations, print_function, unicode_literals

import importlib
import os
import pathlib
from typing import Dict, List, Optional, Union

import cowsay
from hubble import AuthenticationRequiredError
from kubernetes import client, config
from pyfiglet import Figlet

from now.apps.base.app import JinaNOWApp
from now.constants import AVAILABLE_DATASET, Apps, DatasetTypes
from now.deployment.deployment import cmd
from now.log import yaspin_extended
from now.now_dataclasses import UserInput
from now.thirdparty.PyInquirer import Separator
from now.thirdparty.PyInquirer.prompt import prompt
from now.utils import _get_info_hubble, jina_auth_login, sigmap, to_camel_case

cur_dir = pathlib.Path(__file__).parent.resolve()
NEW_CLUSTER = {'name': '🐣 create new', 'value': 'new'}
AVAILABLE_SOON = 'will be available in upcoming versions'


def _configure_app_options(app_instance: JinaNOWApp, user_input, **kwargs):
    for option in app_instance.options:
        val = _prompt_value(
            **option,
            **kwargs,
        )
        setattr(user_input, option['name'], val)
    # _configure_quality(user_input, **kwargs)


def _construct_app(app: Apps) -> JinaNOWApp:
    return getattr(
        importlib.import_module(f'now.apps.{app}.app'),
        f'{to_camel_case(app)}',
    )()


def configure_user_input(**kwargs) -> [JinaNOWApp, UserInput]:
    print_headline()

    user_input = UserInput()
    _configure_app(user_input, **kwargs)
    app_instance = _construct_app(user_input.app)
    _configure_app_options(app_instance, user_input, **kwargs)
    _configure_dataset(app_instance, user_input, **kwargs)
    _configure_cluster(user_input, **kwargs)
    app_instance.run_checks(user_input)
    if user_input.deployment_type == 'remote':
        if _configure_security(user_input, **kwargs):
            if _configure_additional_user(user_input, **kwargs):
                _configure_user_emails(user_input, **kwargs)
    return app_instance, user_input


def print_headline():
    f = Figlet(font='slant')
    print('Welcome to:')
    print(f.renderText('Jina NOW'))
    print('Get your search case up and running - end to end.\n')
    print(
        'You can choose between image and text search. \nJina NOW trains a model, pushes it to Jina Hub'
        'and deploys a Flow and a playground app in the cloud or locally. \nCheck out one of our demos or bring '
        'your own data.\n'
    )
    print('If you want learn more about our framework please visit docs.jina.ai')
    print(
        '💡 Make sure you give enough memory to your Docker daemon. '
        '5GB - 8GB should be okay.'
    )
    print()


def _configure_app(user_input: UserInput, **kwargs) -> None:
    """Asks user questions to set output_modality in user_input"""
    user_input.app = _prompt_value(
        name='app',
        choices=[
            {'name': '📝 ▶ 🏞 text to image search', 'value': Apps.TEXT_TO_IMAGE},
            {'name': '🏞 ▶ 📝 image to text search', 'value': Apps.IMAGE_TO_TEXT},
            {'name': '🏞 ▶ 🏞 image to image search', 'value': Apps.IMAGE_TO_IMAGE},
            {'name': '📝 ▶ 📝 text to text search', 'value': Apps.TEXT_TO_TEXT},
            {
                'name': '📝 ▶ 🎦 text to video search (gif only at the moment)',
                'value': Apps.TEXT_TO_VIDEO,
            },
            {
                'name': '🥁 ▶ 🥁 music to music search',
                'value': Apps.MUSIC_TO_MUSIC,
            },
        ],
        prompt_message='What sort of search engine would you like to build?',
        prompt_type='list',
        **kwargs,
    )


def _configure_dataset(
    app_instance: JinaNOWApp, user_input: UserInput, **kwargs
) -> None:
    """Asks user to set dataset attribute of user_input"""
    _configure_app_dataset(app_instance, user_input, **kwargs)

    if user_input.data in [
        name for name, _ in AVAILABLE_DATASET[app_instance.output_modality]
    ]:
        user_input.is_custom_dataset = False
    else:
        user_input.is_custom_dataset = True
        if user_input.data == 'custom':
            _configure_custom_dataset(user_input, **kwargs)
        else:
            _parse_custom_data_from_cli(user_input)


def _parse_custom_data_from_cli(user_input: UserInput):
    data = user_input.data
    try:
        data = os.path.expanduser(data)
    except Exception:
        pass
    if os.path.exists(data):
        user_input.custom_dataset_type = DatasetTypes.PATH
        user_input.dataset_path = data
    elif 'http' in data:
        user_input.custom_dataset_type = DatasetTypes.URL
        user_input.dataset_url = data
    else:
        user_input.custom_dataset_type = DatasetTypes.DOCARRAY
        user_input.dataset_name = data


def _configure_custom_dataset(user_input: UserInput, **kwargs) -> None:
    """Asks user questions to setup custom dataset in user_input."""
    user_input.custom_dataset_type = _prompt_value(
        name='custom_dataset_type',
        prompt_message='How do you want to provide input? (format: https://docarray.jina.ai/)',
        choices=[
            {
                'name': 'DocArray name (recommended)',
                'value': DatasetTypes.DOCARRAY,
            },
            {
                'name': 'DocArray URL',
                'value': DatasetTypes.URL,
            },
            {
                'name': 'Local path',
                'value': DatasetTypes.PATH,
            },
            {'name': 'Download from S3 bucket', 'value': DatasetTypes.S3_BUCKET},
        ],
        **kwargs,
    )
    if user_input.custom_dataset_type == DatasetTypes.DOCARRAY:
        user_input.dataset_name = _prompt_value(
            name='dataset_name',
            prompt_message='Please enter your DocArray name:',
            **kwargs,
        )

    elif user_input.custom_dataset_type == DatasetTypes.URL:
        user_input.dataset_url = _prompt_value(
            name='dataset_url',
            prompt_message='Please paste in the URL to download your DocArray from:',
            prompt_type='input',
            **kwargs,
        )

    elif user_input.custom_dataset_type == DatasetTypes.PATH:
        user_input.dataset_path = _prompt_value(
            name='dataset_path',
            prompt_message='Please enter the path to the local folder:',
            prompt_type='input',
            **kwargs,
        )

    elif user_input.custom_dataset_type == DatasetTypes.S3_BUCKET:
        user_input.dataset_path = _prompt_value(
            name='dataset_path',
            prompt_message='Please enter the S3 URI to the folder:',
            prompt_type='input',
            **kwargs,
        )
        user_input.aws_access_key_id = _prompt_value(
            name='aws_access_key_id',
            prompt_message='Please enter the AWS access key ID:',
            prompt_type='input',
            **kwargs,
        )
        user_input.aws_secret_access_key = _prompt_value(
            name='aws_secret_access_key',
            prompt_message='Please enter the AWS secret access key:',
            prompt_type='input',
            **kwargs,
        )
        user_input.aws_region_name = _prompt_value(
            name='aws_region_name',
            prompt_message='Please enter the AWS region:',
            prompt_type='input',
            **kwargs,
        )


def _configure_app_dataset(
    app_instance: JinaNOWApp, user_input: UserInput, **kwargs
) -> None:
    user_input.data = _prompt_value(
        name='data',
        prompt_message='What dataset do you want to use?',
        choices=[
            {'name': name, 'value': value}
            for value, name in AVAILABLE_DATASET[app_instance.output_modality]
        ]
        + [
            Separator(),
            {
                'name': '✨ custom',
                'value': 'custom',
            },
        ],
        **kwargs,
    )


def _configure_cluster(user_input: UserInput, skip=False, **kwargs):
    """Asks user question to determine cluster for user_input object"""

    def ask_deployment():
        user_input.deployment_type = _prompt_value(
            name='deployment_type',
            choices=[
                {
                    'name': '⛅️ Jina Cloud',
                    'value': 'remote',
                },
                {
                    'name': '📍 Local',
                    'value': 'local',
                },
            ],
            prompt_message='Where do you want to deploy your search engine?',
            prompt_type='list',
            **kwargs,
        )

    if not skip:
        ask_deployment()

    if user_input.deployment_type == 'remote':
        try:
            jina_auth_login()
        except AuthenticationRequiredError:
            with yaspin_extended(
                sigmap=sigmap, text='Log in to JCloud', color='green'
            ) as spinner:
                cmd('jina auth login')
            spinner.ok('🛠️')

        _get_info_hubble(user_input)
        os.environ['JCLOUD_NO_SURVEY'] = '1'
    else:
        # get all local cluster contexts
        choices = _construct_local_cluster_choices(
            active_context=kwargs.get('active_context'), contexts=kwargs.get('contexts')
        )
        # prompt the local cluster context choices to the user
        user_input.cluster = _prompt_value(
            name='cluster',
            choices=choices,
            prompt_message='On which cluster do you want to deploy your search engine?',
            prompt_type='list',
            **kwargs,
        )
        if user_input.cluster != NEW_CLUSTER['value']:
            if not _cluster_running(user_input.cluster):
                print(
                    f'Cluster {user_input.cluster} is not running. Please select a different one.'
                )
                _configure_cluster(user_input, skip=True, **kwargs)
        else:
            user_input.create_new_cluster = True


def _configure_security(user_input: UserInput, **kwargs):
    user_input.secured = _prompt_value(
        name='secured',
        prompt_message='Do you want to secure the flow?',
        choices=[
            {'name': '✅ yes', 'value': True},
            {'name': '⛔ no', 'value': False},
        ],
        **kwargs,
    )
    return user_input.secured


def _configure_additional_user(user_input: UserInput, **kwargs):
    user_input.additional_user = _prompt_value(
        name='additional_user',
        prompt_message='Do you want to provide additional users access to this flow?',
        choices=[
            {'name': '✅ yes', 'value': True},
            {'name': '⛔ no', 'value': False},
        ],
        **kwargs,
    )
    return user_input.additional_user


def _configure_user_emails(user_input: UserInput, **kwargs):
    user_emails_string = _prompt_value(
        name='user_emails',
        prompt_message='Please enter the comma separated Email IDs'
        ' who will have access to this flow:',
        prompt_type='input',
        **kwargs,
    )
    user_emails = user_emails_string.split(',')
    user_input.user_emails = user_emails


def _construct_local_cluster_choices(active_context, contexts):
    context_names = _get_context_names(contexts, active_context)
    choices = [NEW_CLUSTER]
    # filter contexts with `gke`
    if len(context_names) > 0 and len(context_names[0]) > 0:
        context_names = [context for context in context_names if 'gke' not in context]
        choices = context_names + choices
    return choices


def maybe_prompt_user(questions, attribute, **kwargs):
    """
    Checks the `kwargs` for the `attribute` name. If present, the value is returned directly.
    If not, the user is prompted via the cmd-line using the `questions` argument.

    :param questions: A dictionary that is passed to `PyInquirer.prompt`
        See docs: https://github.com/CITGuru/PyInquirer#documentation
    :param attribute: Name of the value to get. Make sure this matches the name in `kwargs`

    :return: A single value of either from `kwargs` or the user cli input.
    """
    if kwargs and kwargs.get(attribute) is not None:
        return kwargs[attribute]
    else:
        answer = prompt(questions)
        if attribute in answer:
            return answer[attribute]
        else:
            print("\n" * 10)
            cowsay.cow('see you soon 👋')
            exit(0)


def _prompt_value(
    name: str,
    prompt_message: str,
    prompt_type: str = 'input',
    choices: Optional[List[Union[Dict, str]]] = None,
    **kwargs: Dict,
):
    qs = {'name': name, 'type': prompt_type, 'message': prompt_message}

    if choices is not None:
        qs['choices'] = choices
        qs['type'] = 'list'
    return maybe_prompt_user(qs, name, **kwargs)


def _get_context_names(contexts, active_context=None):
    names = [c for c in contexts] if contexts is not None else []
    if active_context is not None:
        names.remove(active_context)
        names = [active_context] + names
    return names


def _cluster_running(cluster):
    config.load_kube_config(context=cluster)
    v1 = client.CoreV1Api()
    try:
        v1.list_namespace()
    except Exception as e:
        return False
    return True



import json
import os
from os.path import expanduser as user
from typing import Dict, Optional

import hubble
from docarray import Document, DocumentArray

from now.apps.base.app import JinaNOWApp
from now.constants import NOW_PREPROCESSOR_VERSION, PREFETCH_NR
from now.data_loading.convert_datasets_to_jpeg import to_thumbnail_jpg
from now.finetuning.run_finetuning import finetune
from now.finetuning.settings import FinetuneSettings, parse_finetune_settings
from now.now_dataclasses import UserInput


def common_get_flow_env_dict(
    finetune_settings: FinetuneSettings,
    encoder_uses: str,
    encoder_with: Dict,
    encoder_uses_with: Dict,
    pre_trained_embedding_size: int,
    indexer_uses: str,
    indexer_resources: Dict,
    user_input: UserInput,
):
    """Returns dictionary for the environments variables for the clip & music flow.yml files."""
    if (
        finetune_settings.perform_finetuning and finetune_settings.bi_modal
    ) or user_input.app == 'music_to_music':
        pre_trained_embedding_size = pre_trained_embedding_size * 2

    config = {
        'ENCODER_NAME': f'jinahub+docker://{encoder_uses}',
        'FINETUNE_LAYER_SIZE': finetune_settings.finetune_layer_size,
        'PRE_TRAINED_EMBEDDINGS_SIZE': pre_trained_embedding_size,
        'INDEXER_NAME': f'jinahub+docker://{indexer_uses}',
        'PREFETCH': PREFETCH_NR,
        'PREPROCESSOR_NAME': f'jinahub+docker://NOWPreprocessor/v{NOW_PREPROCESSOR_VERSION}',
        'APP': user_input.app,
        **encoder_with,
        **indexer_resources,
    }
    if encoder_uses_with.get('pretrained_model_name_or_path'):
        config['PRE_TRAINED_MODEL_NAME'] = encoder_uses_with[
            "pretrained_model_name_or_path"
        ]
    if finetune_settings.perform_finetuning:
        config['FINETUNE_ARTIFACT'] = finetune_settings.finetuned_model_artifact
        config['JINA_TOKEN'] = finetune_settings.token

    # retention days
    if 'NOW_CI_RUN' in os.environ:
        config[
            'RETENTION_DAYS'
        ] = 0  # JCloud will delete after 24hrs of being idle if not deleted in CI
    else:
        config['RETENTION_DAYS'] = 7  # for user deployment set it to 30 days

    return config


def common_setup(
    app_instance: JinaNOWApp,
    user_input: UserInput,
    dataset: DocumentArray,
    encoder_uses: str,
    encoder_uses_with: Dict,
    indexer_uses: str,
    pre_trained_embedding_size: int,
    kubectl_path: str,
    encoder_with: Optional[Dict] = {},
    indexer_resources: Optional[Dict] = {},
) -> Dict:
    # should receive pre embedding size
    finetune_settings = parse_finetune_settings(
        pre_trained_embedding_size=pre_trained_embedding_size,
        user_input=user_input,
        dataset=dataset,
        finetune_datasets=app_instance.finetune_datasets,
    )

    if isinstance(encoder_uses, dict):
        key = 'gpu' if user_input.deployment_type == 'remote' else 'cpu'
        _encoder = encoder_uses[key]
    else:
        _encoder = encoder_uses

    env_dict = common_get_flow_env_dict(
        finetune_settings=finetune_settings,
        encoder_uses=_encoder,
        encoder_with=encoder_with,
        encoder_uses_with=encoder_uses_with,
        pre_trained_embedding_size=pre_trained_embedding_size,
        indexer_uses=indexer_uses,
        indexer_resources=indexer_resources,
        user_input=user_input,
    )

    if finetune_settings.perform_finetuning:
        try:
            artifact_id, token = finetune(
                finetune_settings=finetune_settings,
                app_instance=app_instance,
                dataset=dataset,
                user_input=user_input,
                env_dict=env_dict,
                kubectl_path=kubectl_path,
            )

            finetune_settings.finetuned_model_artifact = artifact_id
            finetune_settings.token = token

            env_dict['FINETUNE_ARTIFACT'] = finetune_settings.finetuned_model_artifact
            env_dict['JINA_TOKEN'] = finetune_settings.token
        except Exception as e:
            print(
                'Finetuning is currently offline. The programm execution still continues without finetuning. Please report the following exception to us:'
            )
            import traceback

            traceback.print_exc()
            finetune_settings.perform_finetuning = False

    app_instance.set_flow_yaml(
        finetuning=finetune_settings.perform_finetuning, dataset_len=len(dataset)
    )

    return env_dict


def preprocess_images(da: DocumentArray) -> DocumentArray:
    """Loads all documents into memory to thumbnail them."""

    def convert_fn(d: Document):
        try:
            if d.tensor is None:
                if d.blob != b'':
                    d.convert_blob_to_image_tensor()
                elif d.uri:
                    d.load_uri_to_image_tensor()
            return to_thumbnail_jpg(d)
        except:
            return d

    for d in da:
        convert_fn(d)
    return DocumentArray(d for d in da if d.blob != b'')


def preprocess_text(da: DocumentArray, split_by_sentences=False) -> DocumentArray:
    """If necessary, loads text for all documents. If asked for, splits documents by sentences."""
    import nltk

    nltk.download('punkt', quiet=True)
    from nltk.tokenize import sent_tokenize

    def convert_fn(d: Document):
        try:
            if not d.text:
                if d.uri:
                    d.load_uri_to_text()
                    d.tags['additional_info'] = d.uri
            return d
        except:
            return d

    def gen_split_by_sentences():
        def _get_sentence_docs(batch):
            ret = []
            for d in batch:
                try:
                    ret += [
                        Document(
                            mime_type='text',
                            text=sentence,
                            tags=d.tags,
                        )
                        for sentence in set(sent_tokenize(d.text.replace('\n', ' ')))
                    ]
                except:
                    pass
            return ret

        for batch in da.map_batch(_get_sentence_docs, backend='process', batch_size=64):
            for d in batch:
                yield d

    for d in da:
        convert_fn(d)

    if split_by_sentences:
        da = DocumentArray(d for d in gen_split_by_sentences())

    return DocumentArray(d for d in da if d.text and d.text != '')


def _get_email():
    try:
        with open(user('~/.jina/config.json')) as fp:
            config_val = json.load(fp)
            user_token = config_val['auth_token']
            client = hubble.Client(token=user_token, max_retries=None, jsonify=True)
            response = client.get_user_info()
        if 'email' in response['data']:
            return response['data']['email']
        return ''
    except FileNotFoundError:
        return ''


def get_indexer_config(num_indexed_samples: int) -> Dict:
    """Depending on the number of samples, which will be indexed, indexer and its resources are determined.

    :param num_indexed_samples: number of samples which will be indexed; should incl. chunks for e.g. text-to-video app
    """
    config = {'indexer_uses': 'AnnLiteNOWIndexer2/0.3.0'}
    threshold1 = 50_000
    threshold2 = 250_000
    if 'NOW_CI_RUN' in os.environ:
        threshold1 = 1_500
    if num_indexed_samples <= threshold1:
        config['indexer_uses'] = 'DocarrayIndexerV2'
        config['indexer_resources'] = {'INDEXER_CPU': 0.1, 'INDEXER_MEM': '2G'}
    elif num_indexed_samples <= threshold2:
        config['indexer_resources'] = {'INDEXER_CPU': 0.1, 'INDEXER_MEM': '2G'}
    else:
        config['indexer_resources'] = {'INDEXER_CPU': 1.0, 'INDEXER_MEM': '4G'}

    return config



from concurrent.futures import ProcessPoolExecutor
from time import time

from now.client import Client
from now.constants import Apps

client = Client(
    jcloud_id='...',
    app=Apps.TEXT_TO_VIDEO,
    api_key='...',
)


def call():
    return client.send_request_bff(
        'text-to-video/search',
        text='girl on a motorbike',
        limit=60,
    )


# Latency Test
start = time()
for i in range(10):
    call()
print(f'Latency: {(time() - start) / 10}s')

# QPS test
start = time()
num_queries = 3000
worker = 100
with ProcessPoolExecutor(max_workers=worker) as executor:
    futures = []

    for i in range(num_queries):
        future = executor.submit(call)
        futures.append(future)
    for future in futures:
        future.result()
print(f'QPS: {num_queries / (time() - start)}s')



import dataclasses
import os
import pathlib
import random
import sys
from time import sleep
from typing import Dict, Optional

from docarray import DocumentArray
from jina.clients import Client

from now.apps.base.app import JinaNOWApp
from now.data_loading.data_loading import load_data
from now.deployment.flow import deploy_flow
from now.log import time_profiler
from now.now_dataclasses import UserInput

cur_dir = pathlib.Path(__file__).parent.resolve()


@time_profiler
def run(app_instance: JinaNOWApp, user_input: UserInput, kubectl_path: str):
    """
    TODO: Write docs

    :param user_input:
    :param kubectl_path:
    :return:
    """
    dataset = load_data(app_instance, user_input)

    env_dict = app_instance.setup(
        dataset=dataset, user_input=user_input, kubectl_path=kubectl_path
    )

    (
        client,
        gateway_host,
        gateway_port,
        gateway_host_internal,
        gateway_port_internal,
    ) = deploy_flow(
        deployment_type=user_input.deployment_type,
        flow_yaml=app_instance.flow_yaml,
        env_dict=env_dict,
        ns='nowapi',
        kubectl_path=kubectl_path,
        secured=user_input.secured,
        admin_emails=user_input.admin_emails,
        user_emails=user_input.user_emails,
    )

    print(f"▶ indexing {len(dataset)} documents")
    params = {
        'user_input': dataclasses.asdict(user_input),
        'traversal_paths': app_instance.index_query_access_paths,
        'access_paths': app_instance.index_query_access_paths,
    }
    if user_input.secured:
        params['jwt'] = user_input.jwt
    call_index(client=client, dataset=dataset, parameters=params, return_results=False)
    print('⭐ Success - your data is indexed')

    return gateway_host, gateway_port, gateway_host_internal, gateway_port_internal


@time_profiler
def call_index(
    client: Client,
    dataset: DocumentArray,
    parameters: Optional[Dict] = None,
    return_results: Optional[bool] = False,
):
    request_size = estimate_request_size(dataset)

    # double check that flow is up and running - should be done by wolf/core in the future
    while True:
        try:
            client.post('/index', inputs=DocumentArray(), parameters=parameters)
            break
        except Exception as e:
            if 'NOW_CI_RUN' in os.environ:
                import traceback

                print(e)
                print(traceback.format_exc())
            sleep(1)

    response = client.post(
        '/index',
        request_size=request_size,
        inputs=dataset,
        show_progress=True,
        parameters=parameters,
    )

    if return_results and response:
        return DocumentArray.from_json(response.to_json())


def estimate_request_size(index):
    if len(index) > 30:
        sample = random.sample(index, 30)
    else:
        sample = index
    size = sum([sys.getsizeof(x.content) for x in sample]) / 30
    max_size = 50_000
    max_request_size = 32
    request_size = max(min(max_request_size, int(max_size / size)), 1)
    return request_size



import base64

from docarray import Document, DocumentArray
from fastapi import HTTPException, status
from jina import Client
from jina.excepts import BadServer


def process_query(
    text: str = '', blob: str = b'', uri: str = None, conditions: dict = None
) -> Document:
    if bool(text) + bool(blob) + bool(uri) != 1:
        raise ValueError(
            f'Can only set one value but have text={text}, blob={blob}, uri={uri}'
        )
    try:
        if uri:
            query_doc = Document(uri=uri)
        elif text:
            query_doc = Document(text=text, mime_type='text')
        elif blob:
            base64_bytes = blob.encode('utf-8')
            message_bytes = base64.decodebytes(base64_bytes)
            query_doc = Document(blob=message_bytes, mime_type='image')
        else:
            raise ValueError('None of the attributes uri, text or blob is set.')
    except BaseException as e:
        raise HTTPException(
            status_code=500,
            detail=f'Not a correct encoded query. Please see the error stack for more information. \n{e}',
        )
    query = {}
    if conditions:
        filter_query = []

        for key, value in conditions.items():
            filter_query.append({f'tags__{key}': {'$eq': value}})
        query = {'$and': filter_query}
    return query_doc, query


def get_jina_client(host: str, port: int) -> Client:
    if 'wolf.jina.ai' in host:
        return Client(host=host)
    else:
        return Client(host=host, port=port)


def jina_client_post(
    data, endpoint: str, inputs, parameters=None, *args, **kwargs
) -> DocumentArray:
    """Posts to the endpoint of the Jina client.

    :param data: contains the request model of the flow
    :param endpoint: endpoint which shall be called, e.g. '/index' or '/search'
    :param inputs: document(s) which shall be passed in
    :param parameters: parameters to pass to the executors, e.g. jwt for securitization or limit for search
    :param args: any additional arguments passed to the `client.post` method
    :param kwargs: any additional keyword arguments passed to the `client.post` method
    :return: response of `client.post`
    """
    if parameters is None:
        parameters = {}
    client = get_jina_client(host=data.host, port=data.port)
    auth_dict = {}
    if data.api_key is not None:
        auth_dict['api_key'] = data.api_key
    if data.jwt is not None:
        auth_dict['jwt'] = data.jwt
    try:
        result = client.post(
            endpoint,
            inputs=inputs,
            parameters={**auth_dict, **parameters},
            *args,
            **kwargs,
        )
    except BadServer as e:
        if 'Not a valid user' in e.args[0].status.description:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='You are not authorised to use this flow',
            )
        else:
            raise e
    return result



import json
import os.path
import pathlib
import tempfile
from os.path import expanduser as user
from time import sleep
from typing import Dict, List

from jina import Flow
from jina.clients import Client
from kubernetes import client as k8s_client
from kubernetes import config
from yaspin.spinners import Spinners

from now.cloud_manager import is_local_cluster
from now.constants import JC_SECRET, NOW_AUTH_EXECUTOR_VERSION
from now.deployment.deployment import apply_replace, cmd, deploy_wolf
from now.log import time_profiler, yaspin_extended
from now.utils import sigmap, write_env_file

cur_dir = pathlib.Path(__file__).parent.resolve()


def batch(data_list, n=1):
    l = len(data_list)
    for ndx in range(0, l, n):
        yield data_list[ndx : min(ndx + n, l)]


def wait_for_lb(lb_name, ns):
    config.load_kube_config()
    v1 = k8s_client.CoreV1Api()
    while True:
        try:
            services = v1.list_namespaced_service(namespace=ns)
            ip = [
                s.status.load_balancer.ingress[0].ip
                for s in services.items
                if s.metadata.name == lb_name
            ][0]
            if ip:
                break
        except Exception:
            pass
        sleep(1)
    return ip


def wait_for_all_pods_in_ns(f, ns, max_wait=1800):
    config.load_kube_config()
    v1 = k8s_client.CoreV1Api()
    for i in range(max_wait):
        pods = v1.list_namespaced_pod(ns).items
        not_ready = [
            'x'
            for pod in pods
            if not pod.status
            or not pod.status.container_statuses
            or not len(pod.status.container_statuses) == 1
            or not pod.status.container_statuses[0].ready
        ]
        if len(not_ready) == 0 and f.num_deployments == len(pods):
            return
        sleep(1)


def deploy_k8s(f, ns, tmpdir, kubectl_path):
    k8_path = os.path.join(tmpdir, f'k8s/{ns}')
    with yaspin_extended(
        sigmap=sigmap, text="Convert Flow to Kubernetes YAML", color="green"
    ) as spinner:
        f.to_k8s_yaml(k8_path)
        spinner.ok('🔄')

    # create namespace
    cmd(f'{kubectl_path} create namespace {ns}')

    # deploy flow
    with yaspin_extended(
        Spinners.earth,
        sigmap=sigmap,
        text="Deploy Jina Flow (might take a bit)",
    ) as spinner:
        gateway_host_internal = f'gateway.{ns}.svc.cluster.local'
        gateway_port_internal = 8080
        if is_local_cluster(kubectl_path):
            apply_replace(
                f'{cur_dir}/k8s_backend-svc-node.yml',
                {'ns': ns},
                kubectl_path,
            )
            gateway_host = 'localhost'
            gateway_port = 31080
        else:
            apply_replace(f'{cur_dir}/k8s_backend-svc-lb.yml', {'ns': ns}, kubectl_path)
            gateway_host = wait_for_lb('gateway-lb', ns)
            gateway_port = 8080
        cmd(f'{kubectl_path} apply -R -f {k8_path}')
        # wait for flow to come up
        wait_for_all_pods_in_ns(f, ns)
        spinner.ok("🚀")
    # work around - first request hangs
    sleep(3)
    return gateway_host, gateway_port, gateway_host_internal, gateway_port_internal


def _extend_flow_yaml(flow_yaml, tmpdir, secured, admin_emails, user_emails):
    if secured:
        if flow_yaml.endswith('.yml') or flow_yaml.endswith('.yaml'):
            with open(flow_yaml, 'r') as f:
                flow_yaml = f.read()
        first_part, second_part = flow_yaml.split('executors:')
        executor_string = f"""
  - name: security_check
    uses: jinahub+docker://AuthExecutor2/{NOW_AUTH_EXECUTOR_VERSION}
    uses_with:
      admin_emails: {admin_emails if admin_emails else []}
      user_emails: {user_emails if user_emails else []}
    jcloud:
      resources:
        memory: 1G
    env:
      JINA_LOG_LEVEL: DEBUG"""
        full_yaml = f'{first_part}executors:{executor_string}{second_part}'
        mod_path = os.path.join(tmpdir, 'mod.yml')
        with open(mod_path, 'w') as f:
            f.write(full_yaml)
        return mod_path
    else:
        return flow_yaml


@time_profiler
def deploy_flow(
    deployment_type: str,
    flow_yaml: str,
    ns: str,
    env_dict: Dict,
    kubectl_path: str,
    secured: bool = False,
    admin_emails: List[str] = None,
    user_emails: List[str] = None,
):
    with tempfile.TemporaryDirectory() as tmpdir:
        flow_yaml = _extend_flow_yaml(
            flow_yaml, tmpdir, secured, admin_emails, user_emails
        )
        env_file = os.path.join(tmpdir, 'dot.env')
        write_env_file(env_file, env_dict)

        if deployment_type == 'remote':
            flow = deploy_wolf(path=flow_yaml, env_file=env_file, name=ns)
            host = flow.gateway
            client = Client(host=host)

            # Dump the flow ID and gateway to keep track
            pathlib.Path(user(JC_SECRET)).parent.mkdir(parents=True, exist_ok=True)
            with open(user(JC_SECRET), 'w') as fp:
                json.dump({'flow_id': flow.flow_id, 'gateway': host}, fp)

            # host & port
            gateway_host = 'remote'
            gateway_port = None
            gateway_host_internal = host
            gateway_port_internal = None  # Since host contains protocol
        else:
            from dotenv import load_dotenv

            load_dotenv(env_file, override=True)
            f = Flow.load_config(flow_yaml)
            (
                gateway_host,
                gateway_port,
                gateway_host_internal,
                gateway_port_internal,
            ) = deploy_k8s(
                f,
                ns,
                tmpdir,
                kubectl_path=kubectl_path,
            )
            client = Client(host=gateway_host, port=gateway_port)

        if os.path.exists(env_file):
            os.remove(env_file)

    return (
        client,
        gateway_host,
        gateway_port,
        gateway_host_internal,
        gateway_port_internal,
    )



import os

from now.apps.base.app import JinaNOWApp
from now.constants import Apps, Modalities


class TextToTextAndImage(JinaNOWApp):
    """
    Hybrid text to text+image search combining symbolic and neural IR approaches.
    """

    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> Apps:
        return Apps.TEXT_TO_TEXT_AND_IMAGE

    @property
    def is_enabled(self) -> bool:
        return False

    @property
    def description(self) -> str:
        return (
            'Text to text+image search app combining symbolic and neural IR approaches.'
        )

    def set_flow_yaml(self, **kwargs):
        """configure the flow yaml in the Jina NOW app."""
        flow_dir = os.path.abspath(os.path.join(__file__, '..'))
        self.flow_yaml = os.path.join(flow_dir, 'flow.yml')

    @property
    def input_modality(self) -> Modalities:
        return Modalities.TEXT

    @property
    def output_modality(self) -> Modalities:
        return Modalities.TEXT_AND_IMAGE



import requests


class Client:
    """
    This is the jina NOW client which can be used to run requests against apps which are deployed using jcloud.
    """

    def __init__(self, jcloud_id, app, api_key):
        self.jcloud_id = jcloud_id
        self.app = app
        self.api_key = api_key

    # def get_maps(self, app_instance, path):
    #     """
    #     checks if the specified path is matching one of the provided patterns
    #     and returns the related request and response mappings
    #     """
    #     for path_regex, (
    #         request_model,
    #         response_model,
    #         request_map,
    #         response_map,
    #     ) in app_instance.bff_mapping_fns.items():
    #         path_compiled_regex = re.compile(fr'{path_regex}')
    #         if path_compiled_regex.match(path):
    #             return request_model, response_model, request_map, response_map
    #     raise ValueError(
    #         f'the path does not match the following patterns: {[path for path, (_, _) in app_instance.bff_mapping_fns.items()]}'
    #     )

    def send_request_bff(self, endpoint: str, **kwargs):
        request_body = {
            "host": f'grpcs://nowapi-{self.jcloud_id}.wolf.jina.ai',
            "api_key": self.api_key,
            **kwargs,
        }
        response = requests.post(
            f'https://nowrun.jina.ai/api/v1/{endpoint}',
            json=request_body,
        )
        return response

    # def send_request(self, endpoint: str, **kwargs):
    #     """
    #     Client to run requests against a deployed flow
    #     """
    #
    #     app_instance = _construct_app(self.app)
    #     request_model, response_model, request_map, response_map = self.get_maps(
    #         app_instance, endpoint
    #     )
    #     app_request = request_model(
    #         host=f'grpcs://nowapi-{self.jcloud_id}.wolf.jina.ai',
    #         api_key=self.api_key,
    #         **kwargs,
    #     )
    #     jina_request = request_map(app_request)
    #     jina_response = jina_client_post(
    #         app_request,
    #         endpoint,
    #         jina_request.data,
    #         jina_request.parameters,
    #         target_executor=r'\Aencoder_clip\Z|\Aindexer\Z',
    #     )
    #     jina_response_model = JinaResponseModel()
    #     jina_response_model.data = jina_response
    #     app_response = response_map(app_request, jina_response_model)
    #     parsed_response = json.loads(parse_obj_as(response_model, app_response).json())
    #     if '__root__' in parsed_response:
    #         parsed_response = parsed_response['__root__']
    #     return parsed_response


# if __name__ == '__main__':
#     client = Client(jcloud_id='...', app=Apps.TEXT_TO_VIDEO, api_key='...')
#     result = client.send_request(
#         '/search',
#         text='girl on a motorbike',
#         limit=60,
#     )
#     print(result)



import base64
import json
import os
import uuid
from copy import deepcopy
from pathlib import Path

from docarray import Document, DocumentArray

from now.apps.base.app import JinaNOWApp
from now.constants import DatasetTypes, DemoDatasets
from now.data_loading.utils import _fetch_da_from_url, get_dataset_url
from now.log import yaspin_extended
from now.now_dataclasses import UserInput
from now.utils import sigmap


def load_data(app: JinaNOWApp, user_input: UserInput) -> DocumentArray:
    """Based on the user input, this function will pull the configured DocArray dataset ready for the preprocessing
    executor.

    :param app: chosen JinaNOWApp
    :param user_input: The configured user object. Result from the Jina Now cli dialog.
    :return: The loaded DocumentArray.
    """
    da = None

    if user_input.is_custom_dataset:
        if user_input.custom_dataset_type == DatasetTypes.DOCARRAY:
            print('⬇  Pull DocArray dataset')
            da = _pull_docarray(user_input.dataset_name)
        elif user_input.custom_dataset_type == DatasetTypes.URL:
            print('⬇  Pull DocArray dataset')
            da = _fetch_da_from_url(user_input.dataset_url)
        elif user_input.custom_dataset_type == DatasetTypes.PATH:
            print('💿  Loading files from disk')
            da = _load_from_disk(app, user_input)
            if any([doc.uri.endswith('.json') for doc in da]):
                da = _load_tags_from_json(da, user_input)
        elif user_input.custom_dataset_type == DatasetTypes.S3_BUCKET:
            da = _list_files_from_s3_bucket(app=app, user_input=user_input)
            if any([doc.uri.endswith('.json') for doc in da]):
                da = _load_tags_from_json(da, user_input)
    else:
        print('⬇  Download DocArray dataset')
        url = get_dataset_url(user_input.data, app.output_modality)
        da = _fetch_da_from_url(url)
        da[:, 'embedding'] = None
    if da is None:
        raise ValueError(
            f'Could not load DocArray dataset. Please check your configuration: {user_input}.'
        )
    if 'NOW_CI_RUN' in os.environ:
        if user_input.data == DemoDatasets.BEST_ARTWORKS:
            da = da[:2000]
        elif user_input.data == DemoDatasets.TUMBLR_GIFS_10K:
            da = da[:600]
        else:
            da = da[:1000]
    da = da.shuffle(seed=42)
    da = deep_copy_da(da)
    return da


def _open_json(path: str):
    with open(path, 'r') as f:
        data = json.load(f)
    return data


def _open_s3_json(path: str, user_input: UserInput):
    import boto3

    path_splits = path.split('/')
    bucket = path_splits[2]
    key = '/'.join(path_splits[3:])

    client = boto3.client(
        's3',
        aws_access_key_id=user_input.aws_access_key_id,
        aws_secret_access_key=user_input.aws_secret_access_key,
    )
    response = client.get_object(Bucket=bucket, Key=key)
    return json.loads(response['Body'].read())


def _load_tags_from_json(da: DocumentArray, user_input: UserInput):

    print('Loading tags!')
    dic = {}
    ids_to_delete = []
    files_in_same_folder = []
    for i, d in enumerate(da):
        folder = d.uri.rsplit('/', 1)[0]
        file_extension = d.uri.split('.')[-1]
        if file_extension in [
            wildcard.split('.')[-1] for wildcard in user_input.app.supported_wildcards
        ]:
            if folder not in dic:
                dic[folder] = i
            else:
                files_in_same_folder.append(i)

    if len(files_in_same_folder) > 0:
        print('Files with the same modality are found within the same folder!')
        print(
            'Printing first 5 ids:',
            ', '.join(str(id) for id in files_in_same_folder[:5]),
        )

    for i, d in enumerate(da):
        folder = d.uri.rsplit('/', 1)[0]
        if d.uri.split('.')[-1] not in [
            wildcard.split('.')[-1] for wildcard in user_input.app.supported_wildcards
        ]:
            ids_to_delete.append(i)
        if d.uri.endswith('.json') and folder in dic:
            if user_input.dataset_path.startswith('s3://'):
                data = _open_s3_json(d.uri, user_input)
            else:
                data = _open_json(d.uri)
            for tag, value in data['tags'].items():
                da[dic[folder]].tags[tag] = value['slug']
    if len(ids_to_delete) > 0:
        del da[ids_to_delete]
    return da


def _pull_docarray(dataset_name: str):
    try:
        return DocumentArray.pull(name=dataset_name, show_progress=True)
    except Exception:
        print(
            '💔 oh no, the secret of your docarray is wrong, or it was deleted after 14 days'
        )
        exit(1)


def _load_to_datauri_and_save_into_tags(d: Document) -> Document:
    d.tags['uri'] = d.uri
    return d.convert_uri_to_datauri()


def _load_from_disk(app: JinaNOWApp, user_input: UserInput) -> DocumentArray:
    dataset_path = user_input.dataset_path.strip()
    if os.path.isfile(dataset_path):
        try:
            return DocumentArray.load_binary(dataset_path)
        except Exception as e:
            print(f'Failed to load the binary file provided under path {dataset_path}')
            exit(1)
    elif os.path.isdir(dataset_path):
        with yaspin_extended(
            sigmap=sigmap, text="Loading data from folder", color="green"
        ) as spinner:
            spinner.ok('🏭')
            docs = DocumentArray.from_files(
                [f"{dataset_path}/{wild_card}" for wild_card in app.supported_wildcards]
            )
            docs.apply(_load_to_datauri_and_save_into_tags)
            return docs
    else:
        raise ValueError(
            f'The provided dataset path {dataset_path} does not'
            f' appear to be a valid file or folder on your system.'
        )


def _list_files_from_s3_bucket(app: JinaNOWApp, user_input: UserInput) -> DocumentArray:
    import boto3.session

    s3_uri = user_input.dataset_path
    if not s3_uri.startswith('s3://'):
        raise ValueError(
            f"Can't process S3 URI {s3_uri} as it assumes it starts with: 's3://'"
        )

    bucket = s3_uri.split('/')[2]
    folder_prefix = '/'.join(s3_uri.split('/')[3:])

    session = boto3.session.Session(
        aws_access_key_id=user_input.aws_access_key_id,
        aws_secret_access_key=user_input.aws_secret_access_key,
    )
    bucket = session.resource('s3').Bucket(bucket)

    docs = []
    with yaspin_extended(
        sigmap=sigmap, text="Listing files from S3 bucket ...", color="green"
    ) as spinner:
        spinner.ok('🏭')
        for obj in list(bucket.objects.filter(Prefix=folder_prefix)):
            if app.supported_wildcards[0] == '**':
                docs.append(Document(uri=f"s3://{bucket.name}/{obj.key}"))
            else:
                for wild_card in app.supported_wildcards:
                    _postfix = wild_card.split('*')[-1]
                    if str(obj.key).endswith(_postfix):
                        docs.append(Document(uri=f"s3://{bucket.name}/{obj.key}"))
                        break

    return DocumentArray(docs)


def _download_from_s3_bucket(app: JinaNOWApp, user_input: UserInput) -> DocumentArray:
    import boto3.session

    s3_uri = user_input.dataset_path
    if not s3_uri.startswith('s3://'):
        raise ValueError(
            f"Can't process S3 URI {s3_uri} as it assumes it starts with: 's3://'"
        )

    data_dir = os.path.expanduser(
        f'~/.cache/jina-now/data/tmp/{base64.b64encode(bytes(s3_uri, "utf-8")).decode("utf-8")}'
    )
    Path(data_dir).mkdir(parents=True, exist_ok=True)

    bucket = s3_uri.split('/')[2]
    folder_prefix = '/'.join(s3_uri.split('/')[3:])

    session = boto3.session.Session(
        aws_access_key_id=user_input.aws_access_key_id,
        aws_secret_access_key=user_input.aws_secret_access_key,
    )
    bucket = session.resource('s3').Bucket(bucket)

    with yaspin_extended(
        sigmap=sigmap, text="Loading data from S3 and creating DocArray", color="green"
    ) as spinner:
        spinner.ok('🏭')

        for obj in list(bucket.objects.filter(Prefix=folder_prefix)):
            if app.supported_wildcards[0] != '**':
                matches_postfix = False
                for wild_card in app.supported_wildcards:
                    _postfix = wild_card.split('*')[-1]
                    if str(obj.key).endswith(_postfix):
                        matches_postfix = True
                        break
                if not matches_postfix:
                    continue
            # create nested directory structure
            path_obj_machine = os.path.join(data_dir, obj.key)
            Path(os.path.dirname(path_obj_machine)).mkdir(parents=True, exist_ok=True)
            # save file with full path locally
            bucket.download_file(obj.key, path_obj_machine)

        docs = DocumentArray.from_files(
            [f"{data_dir}/{wild_card}" for wild_card in app.supported_wildcards]
        )
        docs.apply(_load_to_datauri_and_save_into_tags)
        return docs


def deep_copy_da(da: DocumentArray) -> DocumentArray:
    new_da = DocumentArray()
    for i, d in enumerate(da):
        new_doc = deepcopy(d)
        new_doc.id = str(uuid.uuid4())
        new_da.append(new_doc)
    return new_da



import base64
import io
import os
from typing import Dict, List

import numpy as np
import PIL
from docarray import Document, DocumentArray
from jina.serve.runtimes.gateway.http.models import JinaRequestModel, JinaResponseModel
from now_common.utils import _get_email, common_setup, get_indexer_config
from pydantic import BaseModel

from deployment.bff.app.v1.models.text import NowTextSearchRequestModel
from deployment.bff.app.v1.models.video import (
    NowVideoIndexRequestModel,
    NowVideoListResponseModel,
    NowVideoResponseModel,
)
from now.apps.base.app import JinaNOWApp
from now.constants import CLIP_USES, Apps, Modalities
from now.now_dataclasses import UserInput

NUM_FRAMES_SAMPLED = 3


class TextToVideo(JinaNOWApp):
    def __init__(self):
        super().__init__()

    @property
    def app_name(self) -> str:
        return Apps.TEXT_TO_VIDEO

    @property
    def is_enabled(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return 'Text to video search app'

    @property
    def input_modality(self) -> Modalities:
        return Modalities.TEXT

    @property
    def output_modality(self) -> Modalities:
        return Modalities.VIDEO

    @property
    def required_docker_memory_in_gb(self) -> int:
        return 12

    def set_flow_yaml(self, **kwargs):
        finetuning = kwargs.get('finetuning', False)
        dataset_len = kwargs.get('dataset_len', 0) * NUM_FRAMES_SAMPLED
        is_jina_email = _get_email().split('@')[-1] == 'jina.ai'

        flow_dir = os.path.abspath(os.path.join(__file__, '..'))

        if finetuning:
            if dataset_len > 200_000 and is_jina_email:
                print(f"🚀🚀🚀 You are using high performance flow")
                self.flow_yaml = os.path.join(
                    flow_dir, 'ft-flow-video-clip-high-performance.yml'
                )
            else:
                self.flow_yaml = os.path.join(flow_dir, 'ft-flow-video-clip.yml')
        else:
            if dataset_len > 200_000 and is_jina_email:
                print(f"🚀🚀🚀 You are using high performance flow")
                self.flow_yaml = os.path.join(
                    flow_dir, 'flow-video-clip-high-performance.yml'
                )
            else:
                self.flow_yaml = os.path.join(flow_dir, 'flow-video-clip.yml')

    @property
    def supported_wildcards(self) -> List[str]:
        return ['*.gif', '*.mp4', '*.mov']

    @property
    def index_query_access_paths(self) -> str:
        return '@c'

    def setup(
        self, dataset: DocumentArray, user_input: UserInput, kubectl_path
    ) -> Dict:
        indexer_config = get_indexer_config(len(dataset) * NUM_FRAMES_SAMPLED)
        is_remote = user_input.deployment_type == 'remote'
        encoder_with = {
            'ENCODER_HOST': 'demo-cas.jina.ai' if is_remote else '0.0.0.0',
            'ENCODER_PORT': 2096 if is_remote else None,
            'ENCODER_USES_TLS': True if is_remote else False,
            'ENCODER_IS_EXTERNAL': True if is_remote else False,
        }
        return common_setup(
            app_instance=self,
            user_input=user_input,
            dataset=dataset,
            encoder_uses=CLIP_USES[user_input.deployment_type][0],
            encoder_with=encoder_with,
            encoder_uses_with={
                'pretrained_model_name_or_path': CLIP_USES[user_input.deployment_type][
                    1
                ]
            },
            pre_trained_embedding_size=CLIP_USES[user_input.deployment_type][2],
            indexer_uses=indexer_config['indexer_uses'],
            kubectl_path=kubectl_path,
            indexer_resources=indexer_config['indexer_resources'],
        )

    def preprocess(
        self, da: DocumentArray, user_input: UserInput, is_indexing=False
    ) -> DocumentArray:
        if is_indexing:

            def convert_fn(d: Document):
                try:
                    if d.blob == b'':
                        if d.uri:
                            d.load_uri_to_blob()
                        elif d.tensor is not None:
                            d.convert_tensor_to_blob()
                    sample_video(d)
                except:
                    pass
                return d

            for d in da:
                convert_fn(d)

            return DocumentArray(d for d in da if d.blob != b'')
        else:

            def convert_fn(d: Document):
                d.chunks = DocumentArray(d for d in d.chunks if d.text)
                return d

            for d in da:
                convert_fn(d)

            return DocumentArray(d for d in da if d.chunks)

    @property
    def bff_mapping_fns(self):
        def search_text_to_video_request_mapping_fn(
            request: NowTextSearchRequestModel,
        ) -> JinaRequestModel:
            jina_request_model = JinaRequestModel()
            jina_request_model.data = [Document(chunks=[Document(text=request.text)])]
            jina_request_model.parameters = {
                'limit': request.limit,
                'api_key': request.api_key,
                'jwt': request.jwt,
            }
            return jina_request_model

        def search_video_response_mapping_fn(
            request: NowTextSearchRequestModel, response: JinaResponseModel
        ) -> List[NowVideoResponseModel]:
            docs = response.data
            limit = request.limit
            return docs[0].matches[:limit].to_dict()

        def index_text_to_video_request_mapping_fn(
            request: NowVideoIndexRequestModel,
        ) -> JinaRequestModel:
            index_docs = DocumentArray()
            for video, uri, tags in zip(request.videos, request.uris, request.tags):
                if bool(video) + bool(uri) != 1:
                    raise ValueError(
                        f'Can only set one value but have video={video}, uri={uri}'
                    )
                if video:
                    base64_bytes = video.encode('utf-8')
                    message = base64.decodebytes(base64_bytes)
                    index_docs.append(Document(blob=message, tags=tags))
                else:
                    index_docs.append(Document(uri=uri, tags=tags))
            return JinaRequestModel(data=index_docs)

        def no_response_mapping_fn(_: JinaResponseModel) -> BaseModel:
            return BaseModel()

        return {
            '/search': (
                NowTextSearchRequestModel,
                NowVideoListResponseModel,
                search_text_to_video_request_mapping_fn,
                search_video_response_mapping_fn,
            ),
            '/index': (
                NowVideoIndexRequestModel,
                BaseModel,
                index_text_to_video_request_mapping_fn,
                no_response_mapping_fn,
            ),
        }


def select_frames(num_selected_frames, num_total_frames):
    partition_size = num_total_frames / (num_selected_frames + 1)
    return [round(partition_size * (i + 1)) for i in range(num_selected_frames)]


def sample_video(d):
    video = d.blob
    video_io = io.BytesIO(video)
    gif = PIL.Image.open(video_io)
    frame_indices = select_frames(NUM_FRAMES_SAMPLED, gif.n_frames)
    frames = []
    for i in frame_indices:
        gif.seek(i)
        frame = np.array(gif.convert("RGB"))
        frame_pil = PIL.Image.fromarray(frame)
        frame_pil_resized = frame_pil.resize((224, 224))
        frames.append(frame_pil_resized)
        frame_bytes = io.BytesIO()
        frame_pil_resized.save(frame_bytes, format="JPEG", quality=70)
        d.chunks.append(Document(uri=d.uri, blob=frame_bytes.getvalue()))



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field

_ProtoValueType = Optional[Union[bool, float, str, list, dict]]
_StructValueType = Union[
    _ProtoValueType, List[_ProtoValueType], Dict[str, _ProtoValueType]
]


class _NamedScore(BaseModel):
    value: Optional[float] = None


# Base Request
class BaseRequestModel(BaseModel):
    host: Optional[str] = Field(
        default='localhost', description='Host address returned by the flow deployment.'
    )
    port: Optional[int] = Field(default=31080, description='Port at which to connect.')
    jwt: Optional[Dict[str, Any]] = Field(
        default=None,
        description='User info obtained from the hubble along with token. To be passed when the flow is secure',
    )
    api_key: Optional[str] = Field(
        default=None,
        description='Used to authenticate machines',
    )

    class Config:
        allow_mutation = False
        case_sensitive = False
        arbitrary_types_allowed = True


# Index extending Base Request
class BaseIndexRequestModel(BaseRequestModel):
    tags: List[Dict[str, Any]] = Field(
        default={}, description='List of tags of the documents to be indexed.'
    )
    uris: Optional[List[str]] = Field(
        default=..., description='List of URIs of files or data URIs to index'
    )


# Search extending Base Request
class BaseSearchRequestModel(BaseRequestModel):
    limit: int = Field(default=10, description='Number of matching results to return')
    uri: Optional[str] = Field(
        default=None, description='URI of the file or data URI of the query'
    )
    conditions: Optional[dict] = Field(
        default=None, description='dictionary with filter conditions for search results'
    )


# Base Request for Search
class BaseSearchResponseModel(BaseModel):
    id: str = Field(
        default=..., nullable=False, description='Id of the matching result.'
    )
    scores: Optional[Dict[str, '_NamedScore']] = Field(
        description='Similarity score with respect to the query.'
    )
    tags: Optional[Dict[str, '_StructValueType']] = Field(
        description='Additional tags associated with the file.'
    )
    uri: Optional[str] = Field(
        description='URI of file or data URI of search results', default=''
    )

    class Config:
        case_sensitive = False
        arbitrary_types_allowed = True
